@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_sync.c,v 1.7 2002/08/10 04:32:35 wensong Exp $
+ * Version:     $Id: ip_vs_sync.c,v 1.8 2002/08/17 14:06:02 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -672,6 +672,7 @@ static int sync_thread(void *startup)
 	mm_segment_t oldmm;
 	int state;
 
+	MOD_INC_USE_COUNT;
 	daemonize();
 
 	oldmm = get_fs();
@@ -694,7 +695,10 @@ static int sync_thread(void *startup)
 	mcast_addr.sin_addr.s_addr = htonl(IP_VS_SYNC_GROUP);
 
 	add_wait_queue(&sync_wait, &wait);
+
 	state = ip_vs_sync_state;
+	sync_pid = current->pid;
+	IP_VS_INFO("sync thread started.\n");
 	complete((struct completion *)startup);
 
 	/* processing master/backup loop here */
@@ -707,22 +711,38 @@ static int sync_thread(void *startup)
 	remove_wait_queue(&sync_wait, &wait);
 
 	/* thread exits */
-	stop_sync = 0;
-	wake_up(&stop_sync_wait);
+	sync_pid = 0;
+	IP_VS_INFO("sync thread stopped!\n");
 
 	set_fs(oldmm);
+	MOD_DEC_USE_COUNT;
+
+	stop_sync = 0;
+	wake_up(&stop_sync_wait);
 
 	return 0;
 }
 
+
+static int fork_sync_thread(void *startup)
+{
+	/* fork the sync thread here, then the parent process of the
+	   sync thread is the init process after this thread exits. */
+	if (kernel_thread(sync_thread, startup, 0) < 0)
+		IP_VS_BUG();
+	return 0;
+}
+
+
 int start_sync_thread(int state, char *mcast_ifn)
 {
 	DECLARE_COMPLETION(startup);
+	pid_t pid;
+	int waitpid_result;
 
 	if (sync_pid)
 		return -EEXIST;
 
-	MOD_INC_USE_COUNT;
 	IP_VS_DBG(7, "%s: pid %d\n", __FUNCTION__, current->pid);
 	IP_VS_DBG(7, "Each ip_vs_sync_conn entry need %d bytes\n",
 		  sizeof(struct ip_vs_sync_conn));
@@ -730,11 +750,16 @@ int start_sync_thread(int state, char *mcast_ifn)
 	ip_vs_sync_state = state;
 	strcpy(ip_vs_mcast_ifn, mcast_ifn);
 
-	if ((sync_pid = kernel_thread(sync_thread, &startup, 0)) < 0)
+	if ((pid = kernel_thread(fork_sync_thread, &startup, 0)) < 0)
 		IP_VS_BUG();
+
+	if ((waitpid_result = waitpid(pid, NULL, __WCLONE)) != pid) {
+		IP_VS_ERR("%s: waitpid(%d,...) failed, errno %d\n",
+			  __FUNCTION__, pid, -waitpid_result);
+	}
+
 	wait_for_completion(&startup);
 
-	IP_VS_INFO("sync thread started.\n");
 	return 0;
 }
 
@@ -742,7 +767,6 @@ int start_sync_thread(int state, char *mcast_ifn)
 int stop_sync_thread(void)
 {
 	DECLARE_WAITQUEUE(wait, current);
-	pid_t pid;
 
 	if (!sync_pid)
 		return -ESRCH;
@@ -759,19 +783,11 @@ int stop_sync_thread(void)
 	__set_current_state(TASK_RUNNING);
 	remove_wait_queue(&stop_sync_wait, &wait);
 
-	/* When the same process start/stop the sync thread (in this case
-	   the parent process of the sync thread is this process, instead
-	   of the init process), must reap the sync thread here. */
-	pid = waitpid(sync_pid, NULL, __WCLONE|WNOHANG);
-	if (pid > 0 && pid != sync_pid)
-		IP_VS_ERR("waitpid bug!\n");
+	/* Note: no need to reap the sync thread, because its parent
+	   process is the init process */
 
 	if (stop_sync)
 		IP_VS_BUG();
 
-	IP_VS_INFO("sync thread stopped!\n");
-	sync_pid = 0;
-	MOD_DEC_USE_COUNT;
-
 	return 0;
 }
